// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#ifndef BATTLEFORMATION_H
#define BATTLEFORMATION_H

#include "shiptype.h"

#include "defines.h"

struct Fleet;

// Contains information about a given ship type in a fleet in this battle.
// ShipType + Side + Fleet must be unique in each battle.
struct BattleFormation {
    explicit BattleFormation(const ShipType& shipType,
                             NumberOfShips atStartOfBattle,
                             Fleet* fleet)
        : shipType(shipType)
        , atStartOfBattle(atStartOfBattle)
        , fleet(fleet)
    { }
    implicit BattleFormation(const BattleFormation& other)
        : shipType(other.shipType)
        , atStartOfBattle(other.atStartOfBattle)
        , fleet(other.fleet)
    { }
    implicit BattleFormation(BattleFormation&& other) = delete;
    BattleFormation& operator=(const BattleFormation& other) = delete;
    BattleFormation& operator=(BattleFormation&& other) = delete;

    NumberOfShips kill(NumberOfShips k) {
        // Can't kill ships that have already been killed, stolen or abandoned.
        NumberOfShips killed = std::min(atStartOfBattle - m_stolen - m_killed - m_abandoned, k);
        m_killed += killed;
        return killed;
    }
    NumberOfShips emp(NumberOfShips e) {
        // Can't EMP ships that have already been EMPed.
        NumberOfShips emped = std::min(atStartOfBattle - m_emped, e);
        m_emped += emped;
        return emped;
    }
    NumberOfShips steal(NumberOfShips s) {
        // Can't steal ships that have already been killed, stolen or abandoned.
        NumberOfShips stolen = std::min(atStartOfBattle - m_killed - m_stolen - m_abandoned, s);
        m_stolen += stolen;
        return stolen;
    }
    NumberOfShips abandon(NumberOfShips a) {
        // Can't abandon ships that have already been killed, stolen or
        // abandoned.
        NumberOfShips abandoned = std::min(atStartOfBattle - m_killed - m_stolen - m_abandoned, a);
        m_abandoned += abandoned;
        return abandoned;
    }
    NumberOfShips cap(NumberOfShips c) {
        NumberOfShips capped = std::min(NumberOfShips(0.25 * atStartOfBattle) - m_capped, c);
        m_capped += capped;
        return capped;
    }
    NumberOfShips structureKill(NumberOfShips s) {
        NumberOfShips structureKilled = std::min(NumberOfShips(0.1 * atStartOfBattle) - m_structureKilled, s);
        m_structureKilled += structureKilled;
        return structureKilled;
    }
    Guns gunsReady() const {
        // Abandoned ships have already fired. I think.
        return (atStartOfBattle - m_killed - m_stolen - m_emped) * shipType.guns - m_gunsFired;
    }
    Guns gunsFired() const {
        return m_gunsFired;
    }
    void fireGuns(Guns g) {
        m_gunsFired += g;
    }
    NumberOfShips alive() const {
        return atStartOfBattle - lost();
    }
    NumberOfShips lost() const {
        return m_killed + m_stolen + m_abandoned;
    }
    NumberOfShips eligibleForSalvage() const {
        return m_killed + m_stolen;
    }
    NumberOfShips killed() const {
        return m_killed;
    }
    NumberOfShips unemped() const {
        return atStartOfBattle - m_emped;
    }
    NumberOfShips emped() const {
        return m_emped;
    }
    NumberOfShips stolen() const {
        return m_stolen;
    }
    NumberOfShips abandoned() const {
        return m_abandoned;
    }
    NumberOfShips capped() const {
        return m_capped;
    }
    NumberOfShips uncapped() const {
        return atStartOfBattle - m_capped;
    }
    NumberOfShips structureKilled() const {
        return m_structureKilled;
    }
    NumberOfShips unstructureKilled() const {
        return atStartOfBattle - m_structureKilled;
    }
    std::string toString() const;

    const ShipType& shipType;
    NumberOfShips atStartOfBattle;

    Fleet* fleet;

private:
    NumberOfShips m_emped = 0;
    NumberOfShips m_killed = 0;
    NumberOfShips m_stolen = 0;
    NumberOfShips m_abandoned = 0;
    NumberOfShips m_capped = 0;
    NumberOfShips m_structureKilled = 0;

    NumberOfShips m_gunsFired = 0;
};
bool operator==(const BattleFormation& lhs, const BattleFormation& rhs);
bool operator!=(const BattleFormation& lhs, const BattleFormation& rhs);

#endif /* BATTLEFORMATION_H */
