// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#ifndef SHIPTYPE_H
#define SHIPTYPE_H

#include "defines.h"

#include <string>
#include <array>

#include <cstdint>

typedef int64_t Init;
typedef int64_t Armor;
typedef int64_t Guns;
typedef int64_t ERes;
typedef int64_t Cost;

typedef int64_t XP;
typedef int64_t Score;
typedef int64_t Value;

typedef int64_t NumberOfShips;

typedef int64_t Eff;

// Maximum number of classes ships can fire at in descending order of priority:
// T1, T2, T3, etc.
static const std::size_t MAX_TARGETS = 3;

// Guns may be less (or in theory more) efficient against differently
// prioritized targets.
static const std::array<Eff, MAX_TARGETS> effs { 100, 60, 30 };

struct Class {
    enum C {
        Fighter,
        Corvette,
        Frigate,
        Destroyer,
        Cruiser,
        Battleship,

        None,

        Asteroid,
        Structure
    } clazz;
    implicit Class(C clazz = None)
        : clazz(clazz)
    { }
    explicit Class(const std::string& className)
        : clazz(fromString(className))
    { }
    std::string toString() const;
    C fromString(const std::string& str) const;
    inline bool isShip() const {
        return clazz >= Fighter && clazz <= Battleship;
    }
};
bool operator==(const Class& lhs, const Class& rhs);
bool operator!=(const Class& lhs, const Class& rhs);
bool operator<( const Class& lhs, const Class& rhs);
bool operator>=(const Class& lhs, const Class& rhs);
bool operator<=(const Class& lhs, const Class& rhs);
bool operator>( const Class& lhs, const Class& rhs);

struct GunType {
    enum G {
        // In this bcalc, ships can only fire at other ships. In order to target
        // structures and roids, we classify these ships as well, in order to be
        // able to treat them the same as real ships. Those fake ships need a 'none'
        // gun type.
        None,

        Kill,
        EMP,
        // Cloak is not a gun type but a ship type, and we don't care about ship
        // types in the bcalc. They are normally kill ships and will be treated
        // as such.
        Steal,

        Pod,
        StructureKill
    } type;
    implicit GunType(G type = None)
        : type(type)
    { }
    explicit GunType(const std::string& gunTypeName)
        : type(fromString(gunTypeName))
    { }
    G fromString(const std::string& gunTypeName) const;
    std::string toString() const;
};
bool operator==(const GunType& lhs, const GunType& rhs);
bool operator!=(const GunType& lhs, const GunType& rhs);
bool operator<( const GunType& lhs, const GunType& rhs);
bool operator>=(const GunType& lhs, const GunType& rhs);
bool operator<=(const GunType& lhs, const GunType& rhs);
bool operator>( const GunType& lhs, const GunType& rhs);


struct Race {
    enum R {
        None,
        Ter,
        Cat,
        Xan,
        Zik,
        Etd
    } race;
    implicit Race(R race = None)
        : race(race)
    {
        switch (race) {
        case Cat: { salvageFactor = 1.4; } break;
        default:  { salvageFactor = 1.0; } break;
        }
    }
    explicit Race(const std::string& name)
        : Race(fromString(name))
    { }
    std::string toString() const;
    R fromString(const std::string& name) const;
    double salvageFactor = 1.0;
};
bool operator==(const Race& lhs, const Race& rhs);
bool operator!=(const Race& lhs, const Race& rhs);


// Contains global data from the ship stat listing.
struct ShipType {
    explicit ShipType() {
        m_targets.fill(Class::None);
    }
    explicit ShipType(const std::string& name,
                      Class clazz,
                      const std::array<Class, MAX_TARGETS>& targets,
                      GunType type,
                      Init init,
                      Guns guns,
                      Armor armor,
                      ERes eres,
                      Cost cost,
                      Race race)
        : name(name)
        , clazz(clazz)
        , type(type)
        , init(init)
        , guns(guns)
        , armor(armor)
        , eres(eres)
        , cost(cost)
        , race(race)
        , m_targets(targets)
    { }
    void setTarget(std::size_t index, Class target) {
        m_targets[index] = target;
    }
    inline Class target(std::size_t index) const {
        return m_targets[index];
    }
    // TODO: typedef TargetIndex std::size_t.
    std::size_t targets(Class target) const;
    bool isCombat() const {
        return
            type == GunType::Kill ||
            type == GunType::EMP  ||
            type == GunType::Steal;
    }
    std::string toString() const;

    std::string name = "Uninitialized";
    Class clazz = Class::None;
    GunType type = GunType::None;
    Init init = 0;
    // Historically, kill and steal ships would have 'X guns that deal Y damage
    // each', and both variables could differ from ship to ship. In the current
    // (r30-67+) PA combat engine the damage column holds X, the number of guns,
    // and the damage each gun does is fixed: 1. I chose to use the old method
    // of representing that reality. Let's hope it never changes back, because
    // I've written the bcalc such a way that each gun that fires in the course
    // of a particular battle must deal the exact same amount of damage as every
    // other gun in that battle.
    Guns guns = 0;
    Armor armor = 0;
    ERes eres = 0;
    Cost cost = 0;
    Race race = Race::None;
private:
    std::array<Class, MAX_TARGETS> m_targets = {};
};
bool operator==(const ShipType& lhs, const ShipType& rhs);
bool operator!=(const ShipType& lhs, const ShipType& rhs);

#endif /* SHIPTYPE_H */
