// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "analyzestats.h"

#include "battle.h"

#include "stats.h"
#include "bcalc.h"

#include "logger.h"

#include <numeric>

#include <cmath>

namespace {
typedef std::array<bool, (Class::Battleship + 1 - Class::Fighter)> TargetCoverage;
}

bool operator==(const AnalysisOutcome& lhs, const AnalysisOutcome& rhs) {
    return lhs.landable == rhs.landable &&
        lhs.hasCounter == rhs.hasCounter;
}

bool operator!=(const AnalysisOutcome& lhs, const AnalysisOutcome& rhs) {
    return !(lhs == rhs);
}

TargetCoverage setupBattle(const Stats& stats,
                           Battle& battle,
                           const Fleet& aProto, const Fleet& dProto) {
    Maybe<const BattleFormation> attackPod = aProto.firstPod();

    // Which classes does our attack fleet kill/steal? EMP does not count.
    TargetCoverage hasKill = { false };

    Maybe<Fleet> attack = battle.fleet(Side::Attack, 0);
    attack->race  = aProto.race;
    attack->value = aProto.value;
    attack->score = aProto.value;
    attack->xp    = aProto.xp;
    attack->addBattleFormation(attackPod->shipType, attackPod->atStartOfBattle);
    for (const BattleFormation* bf : aProto.combatShips(attackPod->shipType.clazz)) {
        if (bf->shipType.type == GunType::Kill || bf->shipType.type == GunType::Steal) {
            for (std::size_t index = 0; index < MAX_TARGETS; ++index) {
                Class t = bf->shipType.target(index);
                if (t.isShip()) {
                    hasKill[t.clazz] = true;
                }
            }
        }
        attack->addBattleFormation(bf->shipType, bf->atStartOfBattle);
    }

    Maybe<Fleet> defense = battle.fleet(Side::Defense, 0);
    defense->race  = dProto.race;
    defense->value = dProto.value;
    defense->score = dProto.value;
    defense->xp    = dProto.xp;
    defense->addBattleFormation(*stats["Asteroid"], 800);

    return hasKill;
}

AnalysisOutcome analyzeBattle(const Stats& stats,
                              const Fleet& attacker, const Fleet& defender,
                              double valueRatio,
                              bool useRepeatSameTargetIndex,
                              bool useFlak,
                              Verbosity verbosity) {
    Logger log(verbosity);
    Maybe<const BattleFormation> attackPod  = attacker.firstPod();
    Maybe<const BattleFormation> defensePod = defender.firstPod();
    AnalysisOutcome result;

    auto allAnti = defender.allAnti(attackPod->shipType.clazz);
    result.hasCounter = !allAnti.empty();
    if (result.hasCounter) {
        for (std::size_t index = 0; index <= allAnti.size(); ++index) {
            Battle battle(1, 1);
            std::string defString;
            bool calc = false;
            bool foundFlak = false;
            if (index < allAnti.size()) {
                const BattleFormation* escort = allAnti[index];
                defString = escort->shipType.name;
                TargetCoverage hasKill = setupBattle(stats, battle, attacker, defender);
                battle.addBattleFormation(Side::Defense, 0,
                                          escort->shipType, valueRatio * escort->atStartOfBattle);
                if (useFlak) {
                    for (const BattleFormation& bf : defender.battleFormations) {
                        // Flak are ships in the same class as the designated
                        // defender and ships that are either not targeted at
                        // all, or merely emped. They must also not fire at the
                        // attacking fleet, because then they're the flakkee,
                        // not the flak.
                        if ((bf.shipType.clazz == escort->shipType.clazz || !hasKill[bf.shipType.clazz.clazz]) &&
                            bf.shipType.targets(attackPod->shipType.clazz) >= MAX_TARGETS)
                        {
                            battle.addBattleFormation(Side::Defense, 0,
                                                      bf.shipType, valueRatio * bf.atStartOfBattle);
                            foundFlak = true;
                        }
                    }
                }
                calc = true;
            }
            else if (allAnti.size() > 1) {
                defString = std::accumulate(allAnti.cbegin(), allAnti.cend(), std::string(""),
                                            [](const std::string& str, const BattleFormation* bf) -> std::string {
                                                return str + "+" + bf->shipType.name.c_str();
                                            }).substr(1);
                TargetCoverage hasKill = setupBattle(stats, battle, attacker, defender);
                for (const BattleFormation* defShip : allAnti) {
                    battle.addBattleFormation(Side::Defense, 0,
                                              defShip->shipType, valueRatio * defShip->atStartOfBattle);
                    if (useFlak) {
                        for (const BattleFormation& flak : defender.battleFormations) {
                            // Flak ships may include ships that are already in
                            // allAnti, so exclude them here.
                            if (!battle.battleFormation(Side::Defense, 0, flak.shipType) &&
                                (defShip->shipType.clazz == flak.shipType.clazz || !hasKill[flak.shipType.clazz.clazz]) &&
                                flak.shipType.targets(attackPod->shipType.clazz) >= MAX_TARGETS)
                            {
                                battle.addBattleFormation(Side::Defense, 0,
                                                          flak.shipType, valueRatio * flak.atStartOfBattle);
                                foundFlak = true;
                            }
                        }
                    }
                }
                calc = true;
            }
            // If we need to do a calc with flak, but we have no flak, don't
            // redo the calculation. If we're doing a calc with all defenders,
            // but there's only 1 defender, we've already done that calculation
            // in a calc with 1 defender, so don't redo it.
            if (calc && (!useFlak || foundFlak)) {
                bcalc(battle, useRepeatSameTargetIndex, verbosity);
                log.debug("%s\n",
                          battle.toString().c_str());
                double vulnerability = battle.vulnerable(Side::Attack, verbosity);
                result.vulnerability = std::max(vulnerability, result.vulnerability);
                result.landable = result.vulnerability < 1.0;

                log.info("%s %s (%+7ld) attacking %s %s defending with %2sflakked %-16s(%+7ld) @ %5.2f vuln = %s\n",
                         attackPod->shipType.race.toString().c_str(),
                         attackPod->shipType.clazz.toString().substr(0,2).c_str(),
                         battle.costChange(Side::Attack)/100,
                         defensePod->shipType.race.toString().c_str(),
                         defensePod->shipType.clazz.toString().substr(0,2).c_str(),
                         useFlak ? "" : "un",
                         defString.c_str(),
                         battle.costChange(Side::Defense)/100,
                         vulnerability,
                         vulnerability <= 1.0 ? "Vulnerable" : "Immune");
            }
        }
    }
    return result;
}

void analyzeStats(const Stats& stats,
                  std::vector<Fleet> fleets,
                  bool useRepeatSameTargetIndex,
                  Verbosity verbosity) {
    Logger log(verbosity);
    bool withFlak    = true;
    bool withoutFlak = false;
    if (log.verbosity < Verbosity::Info ) {
        log.error("Roids:");
        for (const Fleet& attack : fleets) {
            log.error("        %s %s",
                     attack.firstPod()->shipType.race.toString().c_str(),
                     attack.firstPod()->shipType.clazz.toString().substr(0,2).c_str());
        }
        log.error("\n");
    }
    for (const Fleet& attack : fleets) {
        Maybe<const BattleFormation> attackPod = attack.firstPod();
        log.info("* %s %s\n",
                 attackPod->shipType.race.toString().c_str(),
                 attackPod->shipType.clazz.toString().c_str());

        if (log.verbosity < Verbosity::Info ) {
            log.error("%s %s",
                     attackPod->shipType.race.toString().c_str(),
                     attackPod->shipType.clazz.toString().substr(0,2).c_str());
        }
        for (const Fleet& defense : fleets) {
            Maybe<const BattleFormation> defensePod = defense.firstPod();
            log.info("** %s %s vs %s %s\n",
                     attackPod->shipType.race.toString().c_str(),
                     attackPod->shipType.clazz.toString().substr(0,2).c_str(),
                     defensePod->shipType.race.toString().c_str(),
                     defensePod->shipType.clazz.toString().substr(0,2).c_str());

            double valueRatio = 0.7;
            AnalysisOutcome result = analyzeBattle(stats,
                                                   attack, defense,
                                                   valueRatio,
                                                   useRepeatSameTargetIndex,
                                                   withoutFlak,
                                                   verbosity);
            AnalysisOutcome intermediate = analyzeBattle(stats,
                                                         attack, defense,
                                                         valueRatio,
                                                         useRepeatSameTargetIndex,
                                                         withFlak,
                                                         verbosity);
            result.landable &= intermediate.landable;
            result.vulnerability = std::max(result.vulnerability, intermediate.vulnerability);
            if (log.verbosity < Verbosity::Warning) {
                log.error("  %12s",
                          result.landable || !result.hasCounter
                          ? "Vulnerable"
                          : "Immume");
            }
            else if (log.verbosity < Verbosity::Info) {
                log.warn("  %5s(%5.2f)",
                         result.landable || !result.hasCounter
                         ? "Vuln"
                         : "Imm",
                         result.vulnerability);
            }
            else {
                if (result.hasCounter) {
                    log.info("%s %s is %-10s to %s %s!\n",
                             defensePod->shipType.race.toString().c_str(),
                             defensePod->shipType.clazz.toString().substr(0,2).c_str(),
                             result.landable
                             ? "vulnerable"
                             : "immune",
                             attackPod->shipType.race.toString().c_str(),
                             attackPod->shipType.clazz.toString().substr(0,2).c_str());
                }
                else {
                    log.info("%s %s does not have a counter to %s %s!\n",
                             defensePod->shipType.race.toString().c_str(),
                             defensePod->shipType.clazz.toString().substr(0,2).c_str(),
                             attackPod->shipType.race.toString().c_str(),
                             attackPod->shipType.clazz.toString().substr(0,2).c_str());
                }
            }
            log.info("\n");
        }
        if (log.verbosity < Verbosity::Info) {
            log.error("\n");
        }
    }
}
