// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "fleet.h"

#include <numeric>
#include <algorithm>

void Fleet::addStolen(const ShipType& shipType, NumberOfShips atStartOfBattle) {
    Maybe<BattleFormation> s = stolen(shipType.name);
    if (s) {
        s->atStartOfBattle += atStartOfBattle;
    }
    else {
        stolenBattleFormations.emplace_back(shipType, atStartOfBattle, this);
    }
}

Maybe<BattleFormation> Fleet::stolen(const std::string& shipTypeName) {
    auto iter = std::find_if(stolenBattleFormations.begin(),
                             stolenBattleFormations.end(),
                             [&shipTypeName](const BattleFormation& bf) -> bool {
                                 return bf.shipType.name == shipTypeName;
                             });
    return Maybe<BattleFormation>(&*iter, iter != stolenBattleFormations.end());
}

Maybe<const BattleFormation> Fleet::stolen(const std::string& shipTypeName) const {
    auto iter = std::find_if(stolenBattleFormations.begin(),
                             stolenBattleFormations.end(),
                             [&shipTypeName](const BattleFormation& bf) -> bool {
                                 return bf.shipType.name == shipTypeName;
                             });
    return Maybe<const BattleFormation>(&*iter, iter != stolenBattleFormations.end());
}

////////////////////////////////////////////////////////////////////////////////

void Fleet::addXP(XP xp) {
    this->xp += xp;
    this->score += 60 * xp;
}

void Fleet::grantSalvage(Cost salvage) {
    this->salvage += salvage * race.salvageFactor;
}

Cost Fleet::fleetCost() const {
    return std::accumulate(battleFormations.cbegin(),
                           battleFormations.cend(),
                           Cost(0),
                           [](Cost cost, const BattleFormation& bf) -> Cost {
                               // Skip ships that are not really ships:
                               // Asteroids and Structures.
                               return cost + (bf.shipType.clazz == Class::Asteroid || bf.shipType.clazz == Class::Structure
                                              ? 0
                                              : bf.shipType.cost * bf.atStartOfBattle);
                           });
}

Cost Fleet::eligibleForSalvage() const {
    return std::accumulate(battleFormations.cbegin(),
                           battleFormations.cend(),
                           Cost(0),
                           [](Cost cost, const BattleFormation& bf) -> Cost {
                               return cost + bf.eligibleForSalvage() * bf.shipType.cost;
                           });
}

Cost Fleet::costStolen() const {
    return std::accumulate(stolenBattleFormations.cbegin(),
                           stolenBattleFormations.cend(),
                           Cost(0),
                           [](Cost cost, const BattleFormation& bf) -> Cost {
                               return cost + (bf.shipType.clazz == Class::Asteroid
                                              ? 0
                                              : bf.atStartOfBattle * bf.shipType.cost);
                           });
}

Cost Fleet::costLost() const {
    return std::accumulate(battleFormations.cbegin(),
                           battleFormations.cend(),
                           Cost(0),
                           [](Cost cost, const BattleFormation& bf) -> Cost {
                               return cost + (bf.shipType.clazz == Class::Asteroid
                                              ? 0
                                              : bf.lost() * bf.shipType.cost);
                           });
}

Cost Fleet::costChange() const {
    return costStolen() + salvage - costLost();
}

////////////////////////////////////////////////////////////////////////////////

Maybe<const BattleFormation> Fleet::battleFormation(const ShipType& shipType) const {
    return battleFormation(shipType.name);
}

Maybe<BattleFormation> Fleet::battleFormation(const ShipType& shipType) {
    return battleFormation(shipType.name);
}

Maybe<const BattleFormation> Fleet::battleFormation(const std::string& shipName) const {
    auto iter = std::find_if(battleFormations.cbegin(),
                             battleFormations.cend(),
                             [&shipName](const BattleFormation& bf) -> bool {
                                 return bf.shipType.name == shipName;
                             });
    return Maybe<const BattleFormation>(&*iter, iter != battleFormations.end());
}

Maybe<BattleFormation> Fleet::battleFormation(const std::string& shipName) {
    auto iter = std::find_if(battleFormations.begin(),
                             battleFormations.end(),
                             [&shipName](const BattleFormation& bf) -> bool {
                                 return bf.shipType.name == shipName;
                             });
    return Maybe<BattleFormation>(&*iter, iter != battleFormations.end());
}

void Fleet::addBattleFormation(const ShipType& shipType, NumberOfShips atStartOfBattle) {
    Maybe<BattleFormation> exist = battleFormation(shipType);
    if (exist) {
        exist->atStartOfBattle += atStartOfBattle;
    }
    else {
        battleFormations.emplace_back(shipType, atStartOfBattle, this);
    }
}

////////////////////////////////////////////////////////////////////////////////

Maybe<const BattleFormation> Fleet::firstPod() const {
    auto iter = std::find_if(battleFormations.cbegin(),
                             battleFormations.cend(),
                             [](const BattleFormation& bf) -> bool {
                                 return bf.shipType.type == GunType::Pod;
                             });
    return Maybe<const BattleFormation>(&*iter, iter != battleFormations.cend());
}

std::vector<const BattleFormation*> Fleet::combatShips(Class clazz) const {
    std::vector<const BattleFormation*> result;
    for (const BattleFormation& bf : battleFormations) {
        if (bf.shipType.type != GunType::Pod &&
            bf.shipType.type != GunType::StructureKill &&
            bf.shipType.clazz == clazz) {
            result.push_back(&bf);
        }
    }
    return result;
}

std::vector<const BattleFormation*> Fleet::allAnti(Class clazz) const {
    std::vector<const BattleFormation*> result;
    for (const BattleFormation& bf : battleFormations) {
        if (bf.shipType.isCombat() && bf.shipType.targets(clazz) < MAX_TARGETS) {
            result.push_back(&bf);
        }
    }
    return result;
}

////////////////////////////////////////////////////////////////////////////////

bool operator==(const Side& lhs, const Side& rhs) {
    return lhs.side == rhs.side;
}

bool operator!=(const Side& lhs, const Side& rhs) {
    return !(lhs == rhs);
}

bool operator<(const Side& lhs, const Side& rhs) {
    return lhs.side < rhs.side;
}

bool operator>=(const Side& lhs, const Side& rhs) {
    return !(lhs < rhs);
}

bool operator<=(const Side& lhs, const Side& rhs) {
    return lhs == rhs || lhs < rhs;
}

bool operator>(const Side& lhs, const Side& rhs) {
    return !(lhs <= rhs);
}

bool operator==(const Fleet& lhs, const Fleet& rhs) {
    return lhs.index == rhs.index
        && lhs.side == rhs.side
        && lhs.score == rhs.score
        && lhs.value == rhs.value
        && lhs.race == rhs.race
        && lhs.salvage == rhs.salvage
        && lhs.battleFormations.size() == rhs.battleFormations.size()
        && lhs.stolenBattleFormations.size() == rhs.stolenBattleFormations.size()
        && std::mismatch(lhs.battleFormations.cbegin(), lhs.battleFormations.cend(),
                         rhs.battleFormations.cbegin(), rhs.battleFormations.cend()).first == lhs.battleFormations.end()
        && std::mismatch(lhs.stolenBattleFormations.cbegin(), lhs.stolenBattleFormations.cend(),
                         rhs.stolenBattleFormations.cbegin(), rhs.stolenBattleFormations.cend()).first == lhs.stolenBattleFormations.end();
}

bool operator!=(const Fleet& lhs, const Fleet& rhs) {
    return !(lhs == rhs);
}
