// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "parser.h"

#include "logger.h"

#include <fstream>
#include <regex>

namespace {

void addBuiltIn(Stats& stats, Verbosity verbosity) {
    Logger log(verbosity);

    Maybe<ShipType> asteroid = stats.create("Asteroid",
                                            Class::Asteroid,
                                            std::array<Class, MAX_TARGETS>{},
                                            GunType::None,
                                            Init(0),
                                            Guns(0),
                                            Armor(50),
                                            ERes(0),
                                            Cost(20000),
                                            Race::None);
    if (asteroid) {
        log.debug("Created Asteroid\n");
    }
    else {
        log.error("Failed to create Asteroid.\n");
    }

    Maybe<ShipType> structure = stats.create("Structure",
                                             Class::Structure,
                                             std::array<Class, MAX_TARGETS>{},
                                             GunType::None,
                                             Init(0),
                                             Guns(0),
                                             Armor(500),
                                             ERes(0),
                                             Cost(20000),
                                             Race::None);
    if (structure) {
        log.debug("Created Structure\n");
    }
    else {
        log.error("Failed to create Structure.\n");
    }
}

} // namespace

Stats readStats(const std::string& statsFile, Verbosity verbosity) {
    Logger log(verbosity);
    Stats stats;

    std::ifstream in(statsFile);
    std::string line(240, '\0');
    std::regex regex("([A-Za-z ]+)\t\
(Fighter|Corvette|Frigate|Destroyer|Cruiser|Battleship)\t\
((?:Fi|Co|Fr|De|Cr|Bs|Ro|St|-|\t)*)\
(Norm|Emp|Cloak|Steal|Pod|Struc)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+|-)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
([0-9]+)\t\
(Ter|Cat|Xan|Zik|Etd)\r?", std::regex::optimize);
    bool ok = true;
    bool done = false;
    while (ok && !done) {
        line.resize(255, '\0');
        in.getline(&line[0], line.size());
        ok = !in.fail();
        done = in.eof();
        if (ok) {
            line.resize(in.gcount() - 1);
            log.trace("Read complete line: '%s'\n",
                      line.c_str());
            // 1:Name	2:Class	3:T1/2/3	4:Type	5:Init	6:Gun	7:Arm	8:Dmg	9:E/R	10:M	11:C	12:E	13:A/C	14:D/C	15:Race
            std::smatch results;
            if (std::regex_match(line, results, regex))
            {
                std::string name(             results[1]);
                Class       clazz(            results[2]);
                std::istringstream targetsStream(results[3]);
                std::vector<std::string> targetStrings(std::istream_iterator<std::string>{targetsStream},
                                                       std::istream_iterator<std::string>{});
                std::array<Class, MAX_TARGETS> targets;
                for (std::size_t i = 0; i < std::min(targetStrings.size(), targets.size()); ++i) {
                    targets[i] = Class(targetStrings[i]);
                }
                for (std::size_t i = targets.size(); i < targetStrings.size(); ++i) {
                    log.error("Application does not support more than %ld target indices, ignoring superfluous target '%s' at index %ld\n",
                              MAX_TARGETS,
                              targetStrings[i].c_str(),
                              i);
                }
                GunType     type(             results[4]);
                Init        init(  std::stoll(results[5]));
                Guns        guns(results[8] == "-"
                                 ? std::stoll(results[6])
                                 : std::stoll(results[8] /* Damage */));
                Armor       armor( std::stoll(results[7]));
                ERes        eres(  std::stoll(results[9]));
                Cost        cost(  std::stoll(results[10]) + std::stoll(results[11]) + std::stoll(results[12]));
                // int64_t  ac(    std::stoll(results[13]));
                // int64_t  dc(    std::stoll(results[14]));
                Race        race(             results[15]);

                log.trace("Name:%s, Class:%s, T:%s/%s/%s, GunType:%s, Init:%ld, Guns:%ld, Armor:%ld, ERes:%ld, Cost:%ld, Race:%s\n",
                          name.c_str(),
                          clazz.toString().c_str(),
                          targets[0].toString().c_str(),
                          targets[1].toString().c_str(),
                          targets[2].toString().c_str(),
                          type.toString().c_str(),
                          init,
                          guns,
                          armor,
                          eres,
                          cost,
                          race.toString().c_str());

                Maybe<ShipType> result = stats.create(name, clazz, targets, type, init, guns, armor, eres, cost, race);

                if (result) {
                    log.debug("Created %s\n",
                              name.c_str());
                }
                else {
                    log.error("Failed to create %s.\n",
                              name.c_str());
                }
            }
            else {
                log.error("Failed to parse line '%s'!\n",
                          line.c_str());
            }
        }
        else if (!done) {
            log.error("Failed to read complete line! Got: '%s'\n",
                      line.c_str());
            log.trace("State: good:%d eof:%d fail:%d bad:%d\n",
                      in.good(),
                      in.eof(),
                      in.fail(),
                      in.bad());
        }
    }
    log.trace("End state: good:%d eof:%d fail:%d bad:%d\n",
              in.good(),
              in.eof(),
              in.fail(),
              in.bad());

    addBuiltIn(stats, verbosity);

    return stats;
}

std::vector<Fleet> readFleets(const Stats& stats, const std::string& file, Verbosity verbosity) {
    std::vector<Fleet> fleets;
    Logger log(verbosity);

    std::ifstream in(file);
    std::string line(240, '\0');
    std::regex battleFormationRegex("\\s+([0-9]+)\\s+([A-Za-z ]+)\r?",
                                    std::regex::optimize);
    std::regex newFleetRegex("([A-Z][a-z]{2})\\s+([A-Z][a-z]):\r?",
                             std::regex::optimize);
    bool ok = true;
    bool done = false;
    Fleet* lastFleet = nullptr;
    while (ok && !done) {
        line.resize(240, '\0');
        in.getline(&line[0], line.size());
        ok = !in.fail();
        done = in.eof();
        if (ok) {
            line.resize(in.gcount() - 1);
            log.trace("Read complete line: '%s'\n", line.c_str());
            // 1:Value	2:Name
            std::smatch results;
            if (std::regex_match(line, results, battleFormationRegex)) {
                Cost cost(std::stoll(results[1]));
                std::string name(results[2]);
                log.trace("Cost:%ld Name:%s\n",
                          cost,
                          name.c_str());
                if (Maybe<const ShipType> ship = stats[name]) {
                    NumberOfShips num = cost / ship->cost;
                    lastFleet->addBattleFormation(*ship, num);
                    log.debug("Added %ld %s to fleet (cost %ld).\n",
                              num,
                              name.c_str(),
                              cost);
                }
                else {
                    log.error("Failed to add %s to fleet: no such ship exists!\n",
                              name.c_str());
                    ok = false;
                }
            }
            else if (std::regex_match(line, results, newFleetRegex)) {
                log.trace("Creating new fleet for %s %s...",
                          results[1].str().c_str(),
                          results[2].str().c_str());
                // Finish the old fleet.
                if (lastFleet) {
                    lastFleet->value = lastFleet->fleetCost() / 90;
                    lastFleet->xp = lastFleet->value / 120;
                    lastFleet->score = lastFleet->value + 60 * lastFleet->xp;
                }
                fleets.emplace_back(0, Side::Defense);
                lastFleet = &fleets.back();
                log.trace(" Done.\n");
            }
            else {
                log.error("Failed to parse line '%s'\n",
                          line.c_str());
            }
        }
        else if (!done) {
            log.error("Failed to read complete line! Got: '%s'\n",
                      line.c_str());
            log.trace("State: good:%d eof:%d fail:%d bad:%d\n",
                      in.good(),
                      in.eof(),
                      in.fail(),
                      in.bad());
        }
    }
    // Finish the last fleet.
    if (lastFleet) {
        lastFleet->value = lastFleet->fleetCost() / 90;
        lastFleet->xp = lastFleet->value / 120;
        lastFleet->score = lastFleet->value + 60 * lastFleet->xp;
    }

    log.trace("End state: good:%d eof:%d fail:%d bad:%d\n",
              in.good(),
              in.eof(),
              in.fail(),
              in.bad());

    return fleets;
}

